-------------------------------------------------------------------------------
-- Title      : wrsw_lj_detect
-- Project    : White Rabbit Switch
-------------------------------------------------------------------------------
-- File       : wrsw_lj_detect.vhd
-- Author     : Mattia Rzzi
-- Company    : CERN BE-CO-HT
-- Platform   : FPGA-generic
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
-- FSM to check the presence of the WRS Low jitter (daughterboard) functionality.
-- Initially, this module was used to detect the LJ daugtherboard, now it is
-- used to detect the LJ functionality, whether integrated onto the main board,
-- or provided on the plugged-in LJ daughterboard.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2012 - 2016 CERN / BE-CO-HT
--
-- This source file is free software; you can redistribute it
-- and/or modify it under the terms of the GNU Lesser General
-- Public License as published by the Free Software Foundation;
-- either version 2.1 of the License, or (at your option) any
-- later version.
--
-- This source is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied
-- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
-- PURPOSE.  See the GNU Lesser General Public License for more
-- details.
--
-- You should have received a copy of the GNU Lesser General
-- Public License along with this source; if not, download it
-- from http://www.gnu.org/licenses/lgpl-2.1.html
--
-------------------------------------------------------------------------------

library ieee;
use ieee.STD_LOGIC_1164.all;
use ieee.numeric_std.all;

use work.wishbone_pkg.all;
use work.gencores_pkg.all;
use work.wr_fabric_pkg.all;
use work.endpoint_pkg.all;
use work.wrsw_top_pkg.all;

library UNISIM;
use UNISIM.vcomponents.all;

entity wrsw_lj_detect is
  generic (
    g_pattern     : std_logic_vector(63 downto 0) := x"CAFED00DCAFED00D";
    g_clk_divider : integer := 16);
  port (
    clk_sys_i        : in  std_logic;
    rst_n_i          : in  std_logic;
    loopback_i       : in  std_logic;
    loopback_o       : out std_logic;
    board_detected_o : out std_logic
  );
end wrsw_lj_detect;

architecture Behavioral of wrsw_lj_detect is

  signal clk_divider    : integer range 0 to g_clk_divider-1;
  signal clk_en         : std_logic;
  signal bit_position   : integer range 0 to g_pattern'length-1;
  signal error_detected : std_logic;

  type fsm_states is (INIT, WRITE_BIT, READ_BIT, DONE);
  signal fsm_state : fsm_states := INIT;

begin

  clock_divider_inst : process (clk_sys_i)
  begin
    if rising_edge (clk_sys_i) then
      clk_divider <= clk_divider + 1;
      if (clk_divider = g_clk_divider - 1) then
        clk_divider <= 0;
        clk_en <= '1';
      else
        clk_en <= '0';
      end if;
    end if;
  end process;

  FSM_INST : process (clk_sys_i)
  begin
    if rising_edge(clk_sys_i) then
      if rst_n_i = '0' then
        board_detected_o <= '0';
        fsm_state        <= INIT;
      else
        case fsm_state is
          when INIT =>
            board_detected_o <= '0';
            bit_position     <= 0;
            fsm_state        <= WRITE_BIT;
            error_detected   <= '0';

          when WRITE_BIT =>
            loopback_o <= g_pattern(bit_position);
            if (clk_en = '1') then
              fsm_state <= READ_BIT;
            end if;

          when READ_BIT =>
            if (g_pattern(bit_position) = loopback_i) then
              if (bit_position = g_pattern'length-1) then
                fsm_state <= DONE;
              else
                bit_position <= bit_position + 1;
                fsm_state    <= WRITE_BIT;
              end if;
            else
              error_detected <= '1';
              fsm_state      <= DONE;
            end if;

          when DONE =>
            board_detected_o <= not error_detected;

          when others =>
            fsm_state <= INIT;
        end case;
      end if;
    end if;
  end process;

end Behavioral;
